#!/usr/bin/env python3
# vim ts=4,fileencoding=utf-8
# SPDXVersion: SPDX-2.3
# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © Copyright 2012-2022, 2023 by Christian Dönges
# SPDXID: SPDXRef-utilities-py
"""Collection of various utility functions.
"""

# pylint: disable=invalid-name

import argparse
import codecs
import contextlib
import ctypes
import glob
import gzip
import locale
import logging
import os
import shutil
import sys
import traceback
from datetime import datetime
from pathlib import Path
from time import mktime
from zipfile import ZipFile, ZipInfo

# Globals (use sparingly)
logger = logging.getLogger(__name__)


def dir_path(path: str) -> Path:
    """Convert the string to a path and return it if it is a directory.

        This function is intended to be used as the `type` argument to :func:`Argparser.add_argument`.

        :param str path: String containing path to check.
        :return: Path to a directory.
        :rtype: Path

        :raise argparse.ArgumentTypeError: The string was not a valid path to a directory.
    """
    if not os.path.isdir(path):
        raise argparse.ArgumentTypeError(
            f'"{path}" is not a valid path to a directory')
    return Path(path)


def exit_hard(returncode: int = 0):
    """Terminate the running application.

        :param int returncode: Exit code to return to spawning shell.
    """
    os._exit(returncode)  # pylint: disable=protected-access


def extract_from_zip_preserving_mtime(zip_file: ZipFile,
                                      zip_info: ZipInfo,
                                      out_path: dir_path):
    """Extract file from a ZIP archive preserving the modification time.

        :param ZipFile zip_file: The archive to extract from.
        :param ZipInfo zip_info: The information about the file to extract.
        :param dir_path out_path: Destination directory for the extracted
            file (must exist).

        :see: https://stackoverflow.com/q/9813243
    """
    zip_file.extract(zip_info, out_path)
    # Add (0, 0, -1) to the time to indicate we do not know if DST was active.
    date_time = mktime(zip_info.date_time + (0, 0, -1))
    # Set ctime and mtime of the extracted file.
    os.utime(out_path / zip_info.filename, (date_time, date_time))


def file_path(path: str) -> Path:
    """Convert the string to a path and return it if it is a file.

        :param str path: String containing path to check.
        :return: Path to a file.
        :rtype: Path

        :raise argparse.ArgumentTypeError: The string did not contain a valid
            path to a file.
    """
    if not os.path.isfile(path):
        raise argparse.ArgumentTypeError(
            f'"{path}" is not a valid path to a file')
    return Path(path)


def get_language() -> str:
    """Determine the language the current user has set their OS to."""
    if os.name == 'posix':
        # BSD, Darwin, and Linux make it easy.
        lang = os.environ['LANG'].split('.')[0]
    elif os.name == 'nt':
        windll = ctypes.windll.kernel32
        lang = locale.windows_locale[windll.GetUserDefaultUILanguage()]
    else:
        raise RuntimeError(f'unknown OS: {os.name}')
    return lang


def hexdump(b: bytes,
            bytes_per_line: int = 16,
            start_offset: int = 0,
            offset_digits: int = 8,
            show_ascii: bool = True) -> str:
    """Generator function to create a pretty representation of the given bytes and return a line per call.

        ```
        first_prefix   00000000  64 65 66 67 68 69 6A 6B  @ABCDEFG
        always_prefix  00000000  64 65 66 67 68 69 6A 6B  @ABCDEFG
        ```

        :param bytes b: The bytes to print.
        :param int bytes_per_line: Number of bytes per line.
        :param int start_offset:   Starting offset for the first byte.
        :param int offset_digits:  Number of digits in the offset.
        :param bool show_ascii: Print ASCII characters after the hex dump if True.
    """
    length = len(b)
    offset = 0
    if length <= bytes_per_line:
        show_offset = False
    else:
        show_offset = True
    while length > 0:
        if length < bytes_per_line:
            bytes_per_line = length
        b_slice = b[offset:offset + bytes_per_line]
        hs = b_slice.hex(' ')
        if show_offset:
            s = f'{offset + start_offset:0{offset_digits}x}  {hs}  '
        else:
            s = f'{hs}  '
        if show_ascii:
            for c in b_slice:
                c = chr(c)
                if c.isprintable():
                    s += c
                else:
                    s += '.'
        length -= bytes_per_line
        offset += bytes_per_line
        yield s


def initialize_console():
    """Initialize the console to work with UTF-8 encoded strings.

        On windows, the console is strange and if output is redirected to a
        file, it gets even stranger. This confuses Python and even though
        PEP 528 solves the problem for interactive consoles, this does not
        help for non-interactive (which redirected streams are).

        The solution for now is to reconfigure the codecs for stdout and
        stderr to always use UTF-8 and replace unmappable characters.
    """
    if os.name == 'nt':
        if sys.stdout.encoding != 'utf-8':
            sys.stdout = codecs.getwriter('utf-8')(sys.stdout.buffer, 'replace')
            print('Reconfigured stdout to use utf-8 encoding.')
        if sys.stderr.encoding != 'utf-8':
            sys.stderr = codecs.getwriter('utf-8')(sys.stderr.buffer, 'replace')
            print('Reconfigured stderr to use utf-8 encoding.')


def logging_add_trace():
    """Add a loglevel TRACE.

        This function should be called only once.
    """
    if hasattr(logging_add_trace, "nr_calls"):
        logging_add_trace.nr_calls += 1
        logging.getLogger(__name__).warning(
            'logging_add_trace() called %d times.',
            logging_add_trace.nr_calls)
        return
    logging_add_trace.nr_calls = 1  # it doesn't exist yet, so initialize it

    if hasattr(logging, 'TRACE'):
        logging.getLogger(__name__).debug('TRACE logging already enabled.')

    # Perform function.
    logging.TRACE = 9
    logging.addLevelName(logging.TRACE, 'TRACE')

    def trace(self, message, *args, **kws):
        """Output message with level TRACE."""
        if self.isEnabledFor(logging.TRACE):
            # Logger takes its '*args' as 'args'.
            self._log(logging.TRACE, message, args, **kws)  # pylint: disable=protected-access

    logging.Logger.trace = trace


def logging_initialize(
        loglevel: int = logging.WARNING,
        log_dir_path: Path = Path('.'),
        log_file_name_format: str = '%P.log',
        log_rotation: int = 0,
        log_compression: int = 5,
        loglevel_console: int = logging.WARNING,
        loglevel_file: int = logging.DEBUG) -> logging.Logger:
    """Initialize the logging interface with the command line options
        passed through the object 'args'. An instance of the root logger is
        returned to the caller.

        If a sub-module uses logging.getLogger('somename'), the logger will
        be a child of the root logger and inherit the settings made here.

        :param int loglevel: Loglevel of the logger. Log messages not meeting
            the level requirement are not processed at all.
        :param Path log_dir_path: Path of the directory containing log files.
            None will prevent log file creation.
        :param str log_file_name_format: Format string for the log file name.
            None will prevent log file creation.
        :param int log_rotation: How many logfiles of the same name to keep.
            If set to 0, any existing log file is overwritten. If set to >0,
            that many old log file copies (named <name>.1, <name>.2, etc.)
            are retained.
        :param int log_compression: Zlib compression level to use in compressing
            log. Level 0 is no compression, level 9 is the highest possible.
        :param int loglevel_console: Loglevel filter of the console logger.
        :param int loglevel_file: Loglevel filter of the file logger.
        :return: The root logger instance for the application.
        :rtype: logging.logger

        :note: If the log_file_name_format contains a timestamp, log_rotation
            will only work on other log file copies with the exact same timestamp.
    """
    # Configure the root logger instance.
    global logger   # # pylint: disable=global-statement
    logger = logging.getLogger()
    logger.setLevel(loglevel)

    # Configure the console handler.
    lc = locale.getpreferredencoding()
    if lc.lower() != 'utf-8':
        locale.setlocale(locale.LC_CTYPE, 'C')
    initialize_console()
    ch = logging.StreamHandler()
    ch.setLevel(loglevel_console)
    terseFormatter = logging.Formatter('%(levelname)-8s %(message)s')
    ch.setFormatter(terseFormatter)
    logger.addHandler(ch)
    logger.debug('Logging to console initialized.')

    if log_file_name_format is not None and log_dir_path is not None:
        log_file_name = string_from_format(log_file_name_format)
        log_file_path = log_dir_path / log_file_name
        rotate_file(log_file_name, log_dir_path, log_rotation, log_compression)

        fh = logging.FileHandler(log_file_path, encoding='utf-8', mode='w')
        fh.setLevel(loglevel_file)
        verboseFormatter = logging.Formatter('%(asctime)s %(name)-26s %(levelname)-8s %(message)s')
        fh.setFormatter(verboseFormatter)
        logger.addHandler(fh)
        logger.debug('Logging to file "%s" initialized.', log_file_path)

    if lc.lower() != 'utf-8':
        if locale.getpreferredencoding().lower() == 'utf-8':
            logger.debug('Changed encoding from "%s" to "utf-8".', lc)
        elif lc == locale.getpreferredencoding():
            logger.debug('Failed to change encoding from "%s" to "utf-8".', lc)
        else:
            logger.warning('Failed to change encoding from "%s" to "utf-8", got "%s".',
                           lc, locale.getpreferredencoding())

    return logger


def initialize_logging(args: argparse.Namespace):
    """Initialize the logging interface with the command line options
        passed through the object 'args'. An instance of the root logger is
        returned to the caller.

        If a sub-module uses `logging.getLogger('somename')`, the logger will
        be a child of the root logger and inherit the settings made here.

        :param argparse.Namespace args: Namespace containing the following
        attributes which are used:
            - `args.debug` bool - Enable or disable debug mode.
            - `args.logging` str - Loglevel (e.g. CRITICAL, ERROR, etc.)
            - `args.verbose` bool - Enable or disable verbose mode.

        :return: The root logger instance for the application.
        :rtype: logging.logger

        .. deprecated:: 1.2.0
           Use :func:`logging_initialize` and :func:`logging_add_trace()` (if
           needed) instead.
    """

    # Define a new log level 'trace'
    logging_add_trace()

    # args.loglevel contains the string value of the command line option
    # --loglevel.
    numeric_log_level = getattr(logging, args.loglevel.upper(), None)
    if not isinstance(numeric_log_level, int):
        raise ValueError(f'Invalid log level: {args.loglevel}')

    # Configure the root logger instance.
    global logger   # # pylint: disable=global-statement
    logger = logging.getLogger()
    if args.debug:
        logger.setLevel(logging.TRACE)
    else:
        logger.setLevel(numeric_log_level)

    # Create file handler only if debug mode is active.
    if args.debug:
        app_name = sys.argv[0]
        fh = logging.FileHandler(app_name + '.log', encoding='utf-8', mode='w')
        fh.setLevel(logging.TRACE)

    lc = locale.getpreferredencoding()
    if lc != 'utf-8':
        locale.setlocale(locale.LC_CTYPE, 'C')

    # Create a console handler.
    initialize_console()
    ch = logging.StreamHandler()
    ch.setLevel(numeric_log_level)

    # Create a formatter and add it to the handlers.
    terseFormatter = logging.Formatter('%(levelname)-8s %(message)s')
    ch.setFormatter(terseFormatter)
    if args.debug:
        verboseFormatter = logging.Formatter(
            '%(asctime)s %(name)-26s %(levelname)-8s %(message)s')
        fh.setFormatter(verboseFormatter)

    # Add the handlers to the logger.
    logger.addHandler(ch)
    if args.debug:
        logger.addHandler(fh)

    if args.verbose:
        logger.info('Logging initialized.')

    if lc != 'utf-8':
        if locale.getpreferredencoding() == 'utf-8':
            logger.debug('Changed encoding from "%s" to "utf-8".', lc)
        elif lc == locale.getpreferredencoding():
            logger.debug('Failed to change encoding from "%s" to "utf-8".', lc)
        else:
            logger.warning('Failed to change encoding from "%s" to "utf-8", got "%s".',
                           lc, locale.getpreferredencoding())

    logger.debug('Parsed args are: %s', args)
    logger.warning('utilities.initialize_logging() is deprecated, use utilities.logging_initialize() instead.')
    return logger


def is_power_of_two(n: int) -> bool:
    """Return True if the given number *n* is a power of two.

        :param int n: number to check
        :return: True if *n* is a power of two, False otherwise.
        :rtype: bool
    """
    return (n != 0) and ((n & (n - 1)) == 0)


def log_hexdump(fn_logger: logging.Logger,
                b: bytes,
                bytes_per_line: int = 16,
                level: int = logging.DEBUG,
                start_offset: int = 0,
                first_prefix: str = '',
                always_prefix: str = '',
                show_ascii: bool = True):
    """Log a pretty representation of the given bytes to the logger.

        ```
        first_prefix   00000000  64 65 66 67 68 69 6A 6B  @ABCDEFG
        always_prefix  00000000  64 65 66 67 68 69 6A 6B  @ABCDEFG
        ```

        :param logging.Logger fn_logger: The logger to log to.
        :param bytes b: The bytes to print.
        :param int bytes_per_line: The number of bytes per line.
        :param int level: Level for logging (e.g. CRITICAL, ERROR, .. DEBUG)
        :param int start_offset: The starting offset for the first byte.
        :param str first_prefix: A string before the offset on the first line.
        :param str always_prefix: A string that will be printed before every
            line (except the first if first_prefix was specified).
        :param bool show_ascii: Print ASCII characters after the hex dump if True.
    """
    if len(first_prefix) == 0:
        first_prefix = always_prefix
    elif len(first_prefix) > len(always_prefix):
        always_prefix += ' ' * (len(first_prefix) - len(always_prefix))
    elif len(always_prefix) > len(first_prefix):
        first_prefix += ' ' * (len(always_prefix) - len(first_prefix))
    first = True
    for line in hexdump(b,
                        bytes_per_line=bytes_per_line,
                        start_offset=start_offset,
                        offset_digits=8,
                        show_ascii=show_ascii):
        if first:
            fn_logger.log(level, '%s  %s', first_prefix, line)
            first = False
        else:
            fn_logger.log(level, '%s  %s', always_prefix, line)


def log_stacktrace(fn_logger: logging.Logger,
                   level: int = logging.DEBUG):
    """Log the current stack trace inside or outside an exception.

        :param logging.Logger fn_logger: The logger to log to.
        :param int level: Level for logging (e.g. CRITICAL, ERROR, .. DEBUG)
    """
    # Get the current exception.
    ex = sys.exc_info()[0]
    # Remove this method from the call stack.
    stack = traceback.extract_stack()[:-1]
    if ex is not None:
        # Exception is present, so remove the call of this method.
        del stack[-1]
    fn_logger.log(level, 'Traceback (most recent call last):')
    for s in traceback.format_list(stack):
        fn_logger.log(level, s)
    if ex is not None:
        fn_logger.log(level, traceback.format_exc())


def resolve_wildcards(filenames: list[str]) -> list[Path]:
    """Resolve unique Paths from a list containing wildcards.

        :param list[str] filenames: A list of filenames that may contain
             wildcards.
        :return: A list of unique Paths.
        :rtype: list[Path]
    """
    # Resolve wildcards.
    names = []
    for n in filenames:
        i = glob.glob(n)
        if i:
            names.extend(i)
    # Remove duplicates
    names = list(set(names))

    # Convert strings to paths.
    paths = [Path(n) for n in names]
    return paths


def rotate_file(file_name: str,
                file_dir: Path,
                rotation: int,
                compress_level: int = 9,
                fn_logger: logging.Logger = logging.getLogger(__name__)):
    """Rotate a (log-)file.

        The parameter rotation specifies the number of copies that will be
        retained.

        When rotating, file_name.rotation-1 is renamed to file_name.rotation
        for all values of rotation down to 1.

        The initial file (with no .rotation suffix) gains a suffix. If
        compress_level > 0, the initial file is also compressed.

        :param str file_name: Name of the log file.
        :param Path file_dir: Path to the directory containing the log file.
        :param int rotation: Number of copies to rotate. 0 to disable.
        :param int compress_level: Zlib compression level with 0 being no
            compression and 9 the highest possible.
        :param logging.Logger fn_logger: Logger used by the function.
    """
    fn_logger.debug('rotate_log(%s, %s, %d)', file_name, file_dir, rotation)
    if rotation == 0:
        fn_logger.debug('rotate_log(): nothing to rotate.')
        return

    max_rotation = rotation
    while rotation >= 0:
        if rotation == 0:
            candidate = file_dir / file_name
            if not candidate.exists():
                fn_logger.debug('rotate_log(): rotation 0 does not exists. Done.')
                return
            candidate_is_compressed = False
        else:
            # Find previously rotated candidate.
            candidate = file_dir / f'{file_name}.{rotation}'
            candidate_is_compressed = False
            if not candidate.exists():
                candidate = file_dir / f'{file_name}.{rotation}.gz'
                candidate_is_compressed = True
                if not candidate.exists():
                    fn_logger.debug('rotate_log(): rotation %d does not exist.', rotation)
                    rotation -= 1
                    continue

        # We have found a match.
        fn_logger.debug('rotate_log(): found candidate %s.', candidate)
        if rotation == 0:
            # All other rotation copies have been renamed, now the last logfile
            # must be compressed (optional) and renamed.
            compress_path = file_dir / file_name
            if compress_level > 0:
                new_name = f'{file_name}.1.gz'
                new_path = file_dir / new_name
                fn_logger.debug('rotate_log(): compress %s -> %s.', compress_path, new_path)
                with open(compress_path, 'rb') as f_in:
                    # Get the modification time of the file we are rotating
                    # so it will be stored correctly in the gzip archive.
                    statinfo = os.stat(compress_path)
                    with open(new_path, 'wb') as f_out:
                        with gzip.GzipFile(file_name,
                                           'wb',
                                           compress_level,
                                           f_out,
                                           mtime=statinfo.st_mtime) as f_zip:
                            shutil.copyfileobj(f_in, f_zip)

                fn_logger.debug('rotate_log(): unlink %s.', compress_path)
                compress_path.unlink()
            else:
                new_path = file_dir / f'{file_name}.{rotation + 1}'
                fn_logger.debug('rotate_log(): rename %s -> %s', candidate, new_path)
                candidate.rename(new_path)
        elif rotation == max_rotation:
            # We have found the maximum rotation file. Delete it to make room.
            fn_logger.debug('rotate_log(): unlinking old %s.', candidate)
            candidate.unlink()
        elif rotation > 0:
            # Rename existing rotation to next higher.
            if candidate_is_compressed:
                new_path = file_dir / f'{file_name}.{rotation + 1}.gz'
            else:
                new_path = file_dir / f'{file_name}.{rotation + 1}'
            fn_logger.debug('rotate_log(): rename %s -> %s', candidate, new_path)
            candidate.rename(new_path)
        rotation -= 1


def round_down(n: int, m: int) -> int:
    """Round the given number *n* down to the nearest multiple of *m*.

        :param int n: number to round
        :param int m: multiple to round to
        :return: n rounded down to a multiple of m.
        :rtype: int
    """
    return n & ~(m - 1)


def round_up(n: int, m: int) -> int:
    """Round the given number *n* up to the nearest multiple of *m*.

        :param int n: number to round
        :param int m: multiple to round to
        :return: n rounded up to a multiple of m.
        :rtype: int
    """
    return (n + m - 1) & ~(m - 1)


@contextlib.contextmanager
def std_open(filename: str = None, mode: str = 'r', encoding: str = 'utf-8'):
    """Open either a file or stdin/stdout for use with `with`.

        If the filename is None or '-' then stdin or stdout (depending on the
        mode) are used.
        Otherwise, the file is used. It is closed when the `with` block is done.

        :param str filename: A filename, '-', or `None`.
        :param str mode: The mode to use for :func:`open`.
        :param str encoding: The encoding to pass to :func:`open`. Defaults to 'utf-8'.

        Example
        -------
        .. code-block:: python

            with std_open(fn, 'r') as f:
                c = f.read()

        .. versionchanged:: 1.3.1
           Default mode changed to 'r' to minimize chances of accidental file clobbering.
    """
    if filename and filename != '-':
        fh = open(filename, mode, encoding=encoding)
    else:
        if 'r' in mode:  # pylint: disable=else-if-used
            fh = sys.stdin
        else:  # 'w'
            fh = sys.stdout

    try:
        yield fh
    finally:
        if fh is not sys.stdin and fh is not sys.stdout:
            fh.close()


def string_from_format(fmt: str) -> str:
    """Given a strftime()-like format, expand into a string.

        Additional formats:
        - %P - name of the application

        :param str fmt: Format string to parse.
        :return: Resulting string.
    """
    fmt2 = ''
    if '%P' in fmt:
        start = 0
        pos = fmt.find('%P', start)
        while pos >= 0:
            length = pos - start
            fmt2 += fmt[start:length] + Path(sys.argv[0]).stem
            start = pos + 2
            pos = fmt.find('%P', start)
        fmt2 += fmt[start:]
    else:
        fmt2 = fmt

    now = datetime.now()
    return now.strftime(fmt2)


def true_stem(path: Path) -> str:
    """Return the true stem (e.g. the name without suffixes) of the Path.

        :param Path path: A Path object.
        :rtype: str
        :return: The final file or directory name without any suffixes.

        Example
        -------

        .. code-block:: python

            print(true_stem(Path('/var/tmp/abc.tar.gz')))
            abc

    """
    ts = path.stem
    while path.suffixes:
        ts = path.stem
        path = Path(ts)
    return ts


if __name__ == '__main__':
    if sys.version_info < (3, 9):
        sys.stderr.write('FATAL ERROR: Python 3.9 or later is required.\n')
        sys.exit(1)



#!/usr/bin/env python3
# vim: fileencoding=utf8
# SPDXVersion: SPDX-2.3
# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © Copyright 2022, 2023 by Christian Dönges
# SPDXID: SPDXRef-littlejonny-py
"""Format output to a text table.

   In case you have been wondering, the name is a reference to
   https://xkcd.com/327/.
"""
import enum
from argparse import Namespace
from dataclasses import dataclass, field
from typing import Any


# Dictionary of select Unicode characters, keyed by name.
UNICODE_CHARACTERS = {
    # Unicode Box Drawing
    # Range: U+2500—U+257F
    'BOX DRAWINGS LIGHT HORIZONTAL': '─',  # U+2500
    'BOX DRAWINGS HEAVY HORIZONTAL': '━',  # U+2501
    'BOX DRAWINGS LIGHT VERTICAL': '│',  # U+2502
    'BOX DRAWINGS HEAVY VERTICAL': '┃',  # U+2503
    'BOX DRAWINGS LIGHT TRIPLE DASH HORIZONTAL': '┄',  # U+2504
    'BOX DRAWINGS HEAVY TRIPLE DASH HORIZONTAL': '┅',  # U+2505
    'BOX DRAWINGS LIGHT TRIPLE DASH VERTICAL': '┆',  # U+2506
    'BOX DRAWINGS HEAVY TRIPLE DASH VERTICAL': '┇',  # U+2507
    'BOX DRAWINGS LIGHT QUADRUPLE DASH HORIZONTAL': '┈',  # U+2508
    'BOX DRAWINGS HEAVY QUADRUPLE DASH HORIZONTAL': '┉',  # U+2509
    'BOX DRAWINGS LIGHT QUADRUPLE DASH VERTICAL': '┊',  # U+250A
    'BOX DRAWINGS HEAVY QUADRUPLE DASH VERTICAL': '┋',  # U+250B
    'BOX DRAWINGS LIGHT DOWN AND RIGHT': '┌',  # U+250C
    'BOX DRAWINGS DOWN LIGHT AND RIGHT HEAVY': '┍',  # U+250D
    'BOX DRAWINGS DOWN HEAVY AND RIGHT LIGHT': '┎',  # U+250E
    'BOX DRAWINGS HEAVY DOWN AND RIGHT': '┏',  # U+250F
    'BOX DRAWINGS LIGHT DOWN AND LEFT': '┐',  # U+2510
    'BOX DRAWINGS DOWN LIGHT AND LEFT HEAVY': '┑',  # U+2511
    'BOX DRAWINGS DOWN HEAVY AND LEFT LIGHT': '┒',  # U+2512
    'BOX DRAWINGS HEAVY DOWN AND LEFT': '┓',  # U+2513
    'BOX DRAWINGS LIGHT UP AND RIGHT': '└',  # U+2514
    'BOX DRAWINGS UP LIGHT AND RIGHT HEAVY': '┕',  # U+2515
    'BOX DRAWINGS UP HEAVY AND RIGHT LIGHT': '┖',  # U+2516
    'BOX DRAWINGS HEAVY UP AND RIGHT': '┗',  # U+2517
    'BOX DRAWINGS LIGHT UP AND LEFT': '┘',  # U+2518
    'BOX DRAWINGS UP LIGHT AND LEFT HEAVY': '┙',  # U+2519
    'BOX DRAWINGS UP HEAVY AND LEFT LIGHT': '┚',  # U+251A
    'BOX DRAWINGS HEAVY UP AND LEFT': '┛',  # U+251B
    'BOX DRAWINGS LIGHT VERTICAL AND RIGHT': '├',  # U+251C
    'BOX DRAWINGS VERTICAL LIGHT AND RIGHT HEAVY': '┝',  # U+251D
    'BOX DRAWINGS UP HEAVY AND RIGHT DOWN LIGHT': '┞',  # U+251E
    'BOX DRAWINGS DOWN HEAVY AND RIGHT UP LIGHT': '┟',  # U+251F
    'BOX DRAWINGS VERTICAL HEAVY AND RIGHT LIGHT': '┠',  # U+2520
    'BOX DRAWINGS DOWN LIGHT AND RIGHT UP HEAVY': '┡',  # U+2521
    'BOX DRAWINGS UP LIGHT AND RIGHT DOWN HEAVY': '┢',  # U+2522
    'BOX DRAWINGS HEAVY VERTICAL AND RIGHT': '┣',  # U+2523
    'BOX DRAWINGS LIGHT VERTICAL AND LEFT': '┤',  # U+2524
    'BOX DRAWINGS VERTICAL LIGHT AND LEFT HEAVY': '┥',  # U+2525
    'BOX DRAWINGS UP HEAVY AND LEFT DOWN LIGHT': '┦',  # U+2526
    'BOX DRAWINGS DOWN HEAVY AND LEFT UP LIGHT': '┧',  # U+2527
    'BOX DRAWINGS VERTICAL HEAVY AND LEFT LIGHT': '┨',  # U+2528
    'BOX DRAWINGS DOWN LIGHT AND LEFT UP HEAVY': '┩',  # U+2529
    'BOX DRAWINGS UP LIGHT AND LEFT DOWN HEAVY': '┪',  # U+252A
    'BOX DRAWINGS HEAVY VERTICAL AND LEFT': '┫',  # U+252B
    'BOX DRAWINGS LIGHT DOWN AND HORIZONTAL': '┬',  # U+252C
    'BOX DRAWINGS LEFT HEAVY AND RIGHT DOWN LIGHT': '┭',  # U+252D
    'BOX DRAWINGS RIGHT HEAVY AND LEFT DOWN LIGHT': '┮',  # U+252E
    'BOX DRAWINGS DOWN LIGHT AND HORIZONTAL HEAVY': '┯',  # U+252F
    'BOX DRAWINGS DOWN HEAVY AND HORIZONTAL LIGHT': '┰',  # U+2530
    'BOX DRAWINGS RIGHT LIGHT AND LEFT DOWN HEAVY': '┱',  # U+2531
    'BOX DRAWINGS LEFT LIGHT AND RIGHT DOWN HEAVY': '┲',  # U+2532
    'BOX DRAWINGS HEAVY DOWN AND HORIZONTAL': '┳',  # U+2533
    'BOX DRAWINGS LIGHT UP AND HORIZONTAL': '┴',  # U+2534
    'BOX DRAWINGS LEFT HEAVY AND RIGHT UP LIGHT': '┵',  # U+2535
    'BOX DRAWINGS RIGHT HEAVY AND LEFT UP LIGHT': '┶',  # U+2536
    'BOX DRAWINGS UP LIGHT AND HORIZONTAL HEAVY': '┷',  # U+2537
    'BOX DRAWINGS UP HEAVY AND HORIZONTAL LIGHT': '┸',  # U+2538
    'BOX DRAWINGS RIGHT LIGHT AND LEFT UP HEAVY': '┹',  # U+2539
    'BOX DRAWINGS LEFT LIGHT AND RIGHT UP HEAVY': '┺',  # U+253A
    'BOX DRAWINGS HEAVY UP AND HORIZONTAL': '┻',  # U+253B
    'BOX DRAWINGS LIGHT VERTICAL AND HORIZONTAL': '┼',  # U+253C
    'BOX DRAWINGS LEFT HEAVY AND RIGHT VERTICAL LIGHT': '┽',  # U+253D
    'BOX DRAWINGS RIGHT HEAVY AND LEFT VERTICAL LIGHT': '┾',  # U+253E
    'BOX DRAWINGS VERTICAL LIGHT AND HORIZONTAL HEAVY': '┿',  # U+253F
    'BOX DRAWINGS UP HEAVY AND DOWN HORIZONTAL LIGHT': '╀',  # U+2540
    'BOX DRAWINGS DOWN HEAVY AND UP HORIZONTAL LIGHT': '╁',  # U+2541
    'BOX DRAWINGS VERTICAL HEAVY AND HORIZONTAL LIGHT': '╂',  # U+2542
    'BOX DRAWINGS LEFT UP HEAVY AND RIGHT DOWN LIGHT': '╃',  # U+2543
    'BOX DRAWINGS RIGHT UP HEAVY AND LEFT DOWN LIGHT': '╄',  # U+2544
    'BOX DRAWINGS LEFT DOWN HEAVY AND RIGHT UP LIGHT': '╅',  # U+2545
    'BOX DRAWINGS RIGHT DOWN HEAVY AND LEFT UP LIGHT': '╆',  # U+2546
    'BOX DRAWINGS DOWN LIGHT AND UP HORIZONTAL HEAVY': '╇',  # U+2547
    'BOX DRAWINGS UP LIGHT AND DOWN HORIZONTAL HEAVY': '╈',  # U+2548
    'BOX DRAWINGS RIGHT LIGHT AND LEFT VERTICAL HEAVY': '╉',  # U+2549
    'BOX DRAWINGS LEFT LIGHT AND RIGHT VERTICAL HEAVY': '╊',  # U+254A
    'BOX DRAWINGS HEAVY VERTICAL AND HORIZONTAL': '╋',  # U+254B
    'BOX DRAWINGS LIGHT DOUBLE DASH HORIZONTAL': '╌',  # U+254C
    'BOX DRAWINGS HEAVY DOUBLE DASH HORIZONTAL': '╍',  # U+254D
    'BOX DRAWINGS LIGHT DOUBLE DASH VERTICAL': '╎',  # U+254E
    'BOX DRAWINGS HEAVY DOUBLE DASH VERTICAL': '╏',  # U+254F
    'BOX DRAWINGS DOUBLE HORIZONTAL': '═',  # U+2550
    'BOX DRAWINGS DOUBLE VERTICAL': '║',  # U+2551
    'BOX DRAWINGS DOWN SINGLE AND RIGHT DOUBLE': '╒',  # U+2552
    'BOX DRAWINGS DOWN DOUBLE AND RIGHT SINGLE': '╓',  # U+2553
    'BOX DRAWINGS DOUBLE DOWN AND RIGHT': '╔',  # U+2554
    'BOX DRAWINGS DOWN SINGLE AND LEFT DOUBLE': '╕',  # U+2555
    'BOX DRAWINGS DOWN DOUBLE AND LEFT SINGLE': '╖',  # U+2556
    'BOX DRAWINGS DOUBLE DOWN AND LEFT': '╗',  # U+2557
    'BOX DRAWINGS UP SINGLE AND RIGHT DOUBLE': '╘',  # U+2558
    'BOX DRAWINGS UP DOUBLE AND RIGHT SINGLE': '╙',  # U+2559
    'BOX DRAWINGS DOUBLE UP AND RIGHT': '╚',  # U+255A
    'BOX DRAWINGS UP SINGLE AND LEFT DOUBLE': '╛',  # U+255B
    'BOX DRAWINGS UP DOUBLE AND LEFT SINGLE': '╜',  # U+255C
    'BOX DRAWINGS DOUBLE UP AND LEFT': '╝',  # U+255D
    'BOX DRAWINGS VERTICAL SINGLE AND RIGHT DOUBLE': '╞',  # U+255E
    'BOX DRAWINGS VERTICAL DOUBLE AND RIGHT SINGLE': '╟',  # U+255F
    'BOX DRAWINGS DOUBLE VERTICAL AND RIGHT': '╠',  # U+2560
    'BOX DRAWINGS VERTICAL SINGLE AND LEFT DOUBLE': '╡',  # U+2561
    'BOX DRAWINGS VERTICAL DOUBLE AND LEFT SINGLE': '╢',  # U+2562
    'BOX DRAWINGS DOUBLE VERTICAL AND LEFT': '╣',  # U+2563
    'BOX DRAWINGS DOWN SINGLE AND HORIZONTAL DOUBLE': '╤',  # U+2564
    'BOX DRAWINGS DOWN DOUBLE AND HORIZONTAL SINGLE': '╥',  # U+2565
    'BOX DRAWINGS DOUBLE DOWN AND HORIZONTAL': '╦',  # U+2566
    'BOX DRAWINGS UP SINGLE AND HORIZONTAL DOUBLE': '╧',  # U+2567
    'BOX DRAWINGS UP DOUBLE AND HORIZONTAL SINGLE': '╨',  # U+2568
    'BOX DRAWINGS DOUBLE UP AND HORIZONTAL': '╩',  # U+2569
    'BOX DRAWINGS VERTICAL SINGLE AND HORIZONTAL DOUBLE': '╪',  # U+256A
    'BOX DRAWINGS VERTICAL DOUBLE AND HORIZONTAL SINGLE': '╫',  # U+256B
    'BOX DRAWINGS DOUBLE VERTICAL AND HORIZONTAL': '╬',  # U+256C
    'BOX DRAWINGS LIGHT ARC DOWN AND RIGHT': '╭',  # U+256D
    'BOX DRAWINGS LIGHT ARC DOWN AND LEFT': '╮',  # U+256E
    'BOX DRAWINGS LIGHT ARC UP AND LEFT': '╯',  # U+256F
    'BOX DRAWINGS LIGHT ARC UP AND RIGHT': '╰',  # U+2570
    'BOX DRAWINGS LIGHT DIAGONAL UPPER RIGHT TO LOWER LEFT': '╱',  # U+2571
    'BOX DRAWINGS LIGHT DIAGONAL UPPER LEFT TO LOWER RIGHT': '╲',  # U+2572
    'BOX DRAWINGS LIGHT DIAGONAL CROSS': '╳',  # U+2573
    'BOX DRAWINGS LIGHT LEFT': '╴',  # U+2574
    'BOX DRAWINGS LIGHT UP': '╵',  # U+2575
    'BOX DRAWINGS LIGHT RIGHT': '╶',  # U+2576
    'BOX DRAWINGS LIGHT DOWN': '╷',  # U+2577
    'BOX DRAWINGS HEAVY LEFT': '╸',  # U+2578
    'BOX DRAWINGS HEAVY UP': '╹',  # U+2579
    'BOX DRAWINGS HEAVY RIGHT': '╺',  # U+257A
    'BOX DRAWINGS HEAVY DOWN': '╻',  # U+257B
    'BOX DRAWINGS LIGHT LEFT AND HEAVY RIGHT': '╼',  # U+257C
    'BOX DRAWINGS LIGHT UP AND HEAVY DOWN': '╽',  # U+257D
    'BOX DRAWINGS HEAVY LEFT AND LIGHT RIGHT': '╾',  # U+257E
    'BOX DRAWINGS HEAVY UP AND LIGHT DOWN': '╿',  # U+257F
}  # UNICODE_CHARACTERS


# Box drawing characters.
# The beauty of Namespace makes it possible to access a character like so:
# box_drawings.light.horizontal.
box_drawings = Namespace(
    double=Namespace(
        down_and_horizontal=UNICODE_CHARACTERS['BOX DRAWINGS DOUBLE DOWN AND HORIZONTAL'],
        down_and_left=UNICODE_CHARACTERS['BOX DRAWINGS DOUBLE DOWN AND LEFT'],
        down_and_right=UNICODE_CHARACTERS['BOX DRAWINGS DOUBLE DOWN AND RIGHT'],
        horizontal=UNICODE_CHARACTERS['BOX DRAWINGS DOUBLE HORIZONTAL'],
        up_and_horizontal=UNICODE_CHARACTERS['BOX DRAWINGS DOUBLE UP AND HORIZONTAL'],
        up_and_left=UNICODE_CHARACTERS['BOX DRAWINGS DOUBLE UP AND LEFT'],
        up_and_right=UNICODE_CHARACTERS['BOX DRAWINGS DOUBLE UP AND RIGHT'],
        vertical=UNICODE_CHARACTERS['BOX DRAWINGS DOUBLE VERTICAL'],
        vertical_and_horizontal=UNICODE_CHARACTERS['BOX DRAWINGS DOUBLE VERTICAL AND HORIZONTAL'],
        vertical_and_left=UNICODE_CHARACTERS['BOX DRAWINGS DOUBLE VERTICAL AND LEFT'],
        vertical_and_right=UNICODE_CHARACTERS['BOX DRAWINGS DOUBLE VERTICAL AND RIGHT'],
    ),  # double
    light=Namespace(
        down_and_horizontal=UNICODE_CHARACTERS['BOX DRAWINGS LIGHT DOWN AND HORIZONTAL'],
        down_and_left=UNICODE_CHARACTERS['BOX DRAWINGS LIGHT DOWN AND LEFT'],
        down_and_right=UNICODE_CHARACTERS['BOX DRAWINGS LIGHT DOWN AND RIGHT'],
        horizontal=UNICODE_CHARACTERS['BOX DRAWINGS LIGHT HORIZONTAL'],
        up_and_horizontal=UNICODE_CHARACTERS['BOX DRAWINGS LIGHT UP AND HORIZONTAL'],
        up_and_left=UNICODE_CHARACTERS['BOX DRAWINGS LIGHT UP AND LEFT'],
        up_and_right=UNICODE_CHARACTERS['BOX DRAWINGS LIGHT UP AND RIGHT'],
        vertical=UNICODE_CHARACTERS['BOX DRAWINGS LIGHT VERTICAL'],
        vertical_and_horizontal=UNICODE_CHARACTERS['BOX DRAWINGS LIGHT VERTICAL AND HORIZONTAL'],
        vertical_and_left=UNICODE_CHARACTERS['BOX DRAWINGS LIGHT VERTICAL AND LEFT'],
        vertical_and_right=UNICODE_CHARACTERS['BOX DRAWINGS LIGHT VERTICAL AND RIGHT'],
    ),  # light
    heavy=Namespace(
        down_and_horizontal=UNICODE_CHARACTERS['BOX DRAWINGS HEAVY DOWN AND HORIZONTAL'],
        down_and_left=UNICODE_CHARACTERS['BOX DRAWINGS HEAVY DOWN AND LEFT'],
        down_and_right=UNICODE_CHARACTERS['BOX DRAWINGS HEAVY DOWN AND RIGHT'],
        horizontal=UNICODE_CHARACTERS['BOX DRAWINGS HEAVY HORIZONTAL'],
        up_and_horizontal=UNICODE_CHARACTERS['BOX DRAWINGS HEAVY UP AND HORIZONTAL'],
        up_and_left=UNICODE_CHARACTERS['BOX DRAWINGS HEAVY UP AND LEFT'],
        up_and_right=UNICODE_CHARACTERS['BOX DRAWINGS HEAVY UP AND RIGHT'],
        vertical=UNICODE_CHARACTERS['BOX DRAWINGS HEAVY VERTICAL'],
        vertical_and_horizontal=UNICODE_CHARACTERS['BOX DRAWINGS HEAVY VERTICAL AND HORIZONTAL'],
        vertical_and_left=UNICODE_CHARACTERS['BOX DRAWINGS HEAVY VERTICAL AND LEFT'],
        vertical_and_right=UNICODE_CHARACTERS['BOX DRAWINGS HEAVY VERTICAL AND RIGHT'],
    )  # heavy
)  # box_drawings


@enum.unique
class LineStyle(enum.Enum):
    """Rendering style for lines."""
    double = box_drawings.double
    light = box_drawings.light
    heavy = box_drawings.heavy


Text = list[str]


@dataclass
class ColumnTable:
    """Table with data arranged by column."""
    _headings: list[str] = field(default_factory=list)  # row or column heading
    _formats: list[str] = field(default_factory=list)  # cell formats
    _cells: list[list[Any]] = field(default_factory=list)  # column[row]
    _num_columns: int = 0
    _num_rows: int = 0

    @property
    def cell_formats(self) -> list[str]:
        """Return headings."""
        return self._formats

    @property
    def headings(self) -> list[str]:
        """Return headings."""
        return self._headings

    @property
    def num_columns(self):
        return self._num_columns

    @property
    def num_rows(self):
        return self._num_rows

    def set_table(self,
                  headings: list[str],
                  cell_formats: list[str],
                  cells: list[list[str]]):
        """Set the headings, the format of the cells, and the cell content.

           The cells are specified column[row].
        """
        if len(headings) != len(cell_formats):
            raise ValueError('number columns in headings and formats do not match')
        if len(headings) != len(cells):
            raise ValueError('number of columns in headings and cells do not match')
        self._num_columns = len(headings)
        self._num_rows = len(cells[0])
        self._headings = headings
        self._formats = cell_formats
        self._cells = cells

    def set_table_transposed(self,
                             headings: list[str],
                             cell_formats: list[str],
                             cells: list[list[str]]):
        """The cells are specified row[column]."""
        transposed_cells = [list(x) for x in zip(*cells)]
        self.set_table(headings, cell_formats, transposed_cells)

    def draw(self, style: LineStyle) -> Text:
        """Draw the table to a list of lines."""
        formatted_cells = []
        column_widths = []
        for c in range(self.num_columns):
            column_width = len(self._headings[c])
            column_data = self._cells[c]
            fmt = self._formats[c]
            formatted_column = []
            for cell in column_data:
                formatted_cell = f'{cell:{fmt}}'
                formatted_column.append(formatted_cell)
                column_width = max(column_width,  len(formatted_cell))
            formatted_cells.append(formatted_column)
            column_widths.append(column_width)

        box = style.value
        lines = []
        first = True
        s = ''
        for c in range(self.num_columns):
            column_width = column_widths[c]
            if first:
                s = box.down_and_right + box.horizontal * (column_width + 2)
                first = False
            else:
                s += box.down_and_horizontal + box.horizontal * (column_width + 2)
        s += box.down_and_left
        lines.append(s)

        s = ''
        for c in range(self.num_columns):
            column_width = column_widths[c]
            s += box.vertical + f' {self._headings[c]:{column_width}s} '
        s += box.vertical
        lines.append(s)

        first = True
        s = ''
        for c in range(self.num_columns):
            column_width = column_widths[c]
            if first:
                s = box.vertical_and_right + box.horizontal * (column_width + 2)
                first = False
            else:
                s += box.vertical_and_horizontal + box.horizontal * (column_width + 2)
        s += box.vertical_and_left
        lines.append(s)

        for r in range(self.num_rows):
            s = ''
            for c in range(self.num_columns):
                column_width = column_widths[c]
                s += box.vertical + f' {formatted_cells[c][r]:{column_width}s} '
            s += box.vertical
            lines.append(s)

        first = True
        s = ''
        for c in range(self.num_columns):
            column_width = column_widths[c]
            if first:
                s = box.up_and_right + box.horizontal * (column_width + 2)
                first = False
            else:
                s += box.up_and_horizontal + box.horizontal * (column_width + 2)
        s += box.up_and_left
        lines.append(s)

        return lines


def draw_box(x, y, style) -> Text:
    """Construct a box of size x * y with style."""
    box = style.value

    lines = []
    s = box.down_and_right + box.horizontal * x + box.down_and_left
    lines.append(s)
    s = box.vertical + ' ' * x + box.vertical
    for i in range(y):
        lines.append(s)
    s = box.up_and_right + box.horizontal * x + box.up_and_left
    lines.append(s)

    return lines


def print_lines(lines: Text):
    """Print lines stored as a list of strings to stdout."""
    for line in lines:
        print(line)


if __name__ == '__main__':
    print_lines(draw_box(5, 3, LineStyle.light))
    print_lines(draw_box(1, 1, LineStyle.double))
    print_lines(draw_box(40, 3, LineStyle.heavy))

    ct1 = ColumnTable()
    ct1.set_table_transposed(
        headings=['First', 'Second', 'Third', 'Fourth'],
        cell_formats=['3d', '08x', 's', '>s'],
        cells=[[1, 0x12345678, 'abc def geh', 'this is a sample text'],
               [2, 0xffffee01, '12345', 'Short'],
               [12, -1, 'minus one', 'negative hex number'],
               [123, 0, 'zero', 'Zero hexadecimal number'],
               [1000, 23, 'First', 'The first column is too large.']]
    )
    print_lines(ct1.draw(LineStyle.light))



#!/usr/bin/env python3
# vim ts=4,fileencoding=utf-8
# SPDX-License-Identifier: Apache-2.0
# SPDXID: pymisclib-1
# PackageCopyrightText: © Copyright 2012-2023 by Christian Dönges <cd.pymisclib@platypus-projects.net >
# PackageDownloadLocation: None
# PackageHomePage: https://rot.regenbogenkraut.de/cd/pymisclib
# PackageName: pymisclib
# PackageOriginator: Originator: Platypus Projects GmbH
# PackageSourceInfo: <text>uses pymisclib from https://rot.regenbogenkraut.de/cd/pymisclib.</text>
# PackageSupplier: Christian Dönges (cd.pymisclib@platypus-projects.net )
# PackageVersion: 1.3.0

from .__version__ import (
    __author__,
    __author_email__,
    __copyright__,
    __description__,
    __license__,
    __title__,
    __url__,
    __version__,
)



#!/usr/bin/env python3
# vim: fileencoding=utf8
# SPDXVersion: SPDX-2.3
# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © Copyright 2012-2022, 2023 by Christian Dönges
# SPDXID: SPDXRef-multiprogressbar-py
"""Print multiple progress bars to the console.

   The configuration of the bars is handled through a list of dictionaries.
   Each dictionary contains the parameters for a single progress bar.

   Before the bars are shown, ``prepare_bars()`` must be called to make room for
   the bars on the console.

   Each time a bar is updated, ``print_bars()`` is called to update the console
   output.

   Bar example
   -----------

   A simple example shows a bar for the entire process and a second bar for
   individual progress steps below the first bar.

   This is an example: ::

       Total |------------------------------| 0% Writing
             |███████████████---------------| 50.9% /tmp3muil9sc/0000000000000000.tmp


   .. code-block:: python

       from multiprogressbar import prepare_bars, print_bars

       config = [
           {
               'current_iteration': 0,
               'total_iterations': 5,
               'prefix': 'Total',
               'suffix': 'Writing',
               'decimals': 0,
               'bar_length': 30},
           {
               'current_iteration': 509,
               'total_iterations': 1000,
               'prefix': '     ',
               'suffix': '/tmp3muil9sc/0000000000000000.tmp',
               'decimals': 1,
               'bar_length': 30},
       ]
       prepare_bars(config)
       print_bars(config)

"""

import enum
import sys
from time import sleep


@enum.unique
class BAR_COLORS(enum.Enum):
    """Colors and the ANSI format strings used to create them."""
    Black = '\033[30m'
    Red = '\033[31m'
    Green = '\033[32m'
    Yellow = '\033[33m'
    Blue = '\033[34m'
    Magenta = '\033[35m'
    Cyan = '\033[36m'
    White = '\033[37m'
    BrightBlack = '\033[1;30m'
    BrightRed = '\033[1;31m'
    BrightGreen = '\033[1;32m'
    BrightYellow = '\033[1;33m'
    BrightBlue = '\033[1;34m'
    BrightMagenta = '\033[1;35m'
    BrightCyan = '\033[1;36m'
    BrightWhite = '\033[1;37m'


def print_single_bar(
        current_iteration: int,
        total_iterations: int,
        prefix: str = '',
        suffix: str = '',
        decimals: int = 1,
        bar_length: int = 80,
        bar_color: BAR_COLORS = BAR_COLORS.Black):
    """Output a single progress bar to stdout.

    :param int current_iteration: the current iteration
    :param int total_iterations: the total number of iterations
    :param str prefix: a string that will be output before the bar
    :param str suffix: a string that will be output after the bar
    :param int bar_length: the length of the bar in characters
    :param BAR_COLORS bar_color: the color of the bar.

    """
    current_iteration = min(current_iteration, total_iterations)

    percents = f'{100 * (current_iteration / float(total_iterations)):.{decimals}f}'
    filled_length = int(round(bar_length * current_iteration / float(total_iterations)))
    the_bar = f'{"█" * filled_length}{"-" * (bar_length - filled_length)}'

    sys.stdout.write(
        f'\x1b[2K\r{prefix} |{bar_color.value}{the_bar}'
        f'{BAR_COLORS.Black.value}| {percents}% {suffix}')


def prepare_bars(configs: list):
    """Print to prepare for the bars."""
    for c in configs:
        sys.stdout.write('\n')


def print_bars(configs: list):
    """Print progress bars."""
    # Move the cursor up to the start of the line of the first bar.
    for n in range(len(configs)):
        sys.stdout.write('\033[F')  # up

    for config in configs:
        ci = config['current_iteration']
        ti = config['total_iterations']
        prefix = config.get('prefix', '')
        suffix = config.get('suffix', '') + '\033[K\n'
        decimals = config.get('decimals', 1)
        bar_length = config.get('bar_length', 80)
        bar_color = config.get('bar_color', BAR_COLORS.Black)
        print_single_bar(ci, ti, prefix, suffix, decimals, bar_length,
                         bar_color)
    sys.stdout.flush()


def main():
    """Demonstration of how it works."""
    configs = [
        {
            'current_iteration': 0,
            'total_iterations': 5,
            'prefix': 'Total',
            'suffix': 'Writing',
            'decimals': 0,
            'bar_length': 30,
            'bar_color': BAR_COLORS.Green,
        },
        {
            'current_iteration': 509,
            'total_iterations': 1000,
            'prefix': '     ',
            'suffix': '/tmp3muil9sc/0000000000000000.tmp',
            'decimals': 1,
            'bar_length': 30
        },
        {
            'current_iteration': 1,
            'total_iterations': 10,
            'bar_length': 15,
            'bar_color': BAR_COLORS.BrightRed,
        },
        {
            'current_iteration': 0,
            'total_iterations': 100,
            'prefix': 'Jolly'
        },
    ]
    prepare_bars(configs)
    for i in range(100):
        print_bars(configs)
        for config in configs:
            config['current_iteration'] += 1
        sleep(.1)


if __name__ == '__main__':
    main()
